import numpy as np
import torch
import sys
import scipy.io
import pdb
import os
import datetime
import scipy.stats as stats
import torch.nn as nn
import torch.optim as optim
from data_loader import VideoDataset
from torch.utils.data import DataLoader
from aim_module import AIM_inserted
from tensorboardX import SummaryWriter
from loss_function import My_loss

import opts
from opts import *

torch.manual_seed(randomseed); torch.cuda.manual_seed_all(randomseed); np.random.seed(randomseed)
torch.backends.cudnn.deterministic = True

def save_model(model, model_name, epoch, path):
    model_path = os.path.join(path, '%s_%d.pth' % (model_name, epoch))
    torch.save(model.state_dict(), model_path)


def train_phase(train_dataloader, optimizer, criterion, epoch):
    model_i3d_lstm.train()   ##  model 
    iter_cnt = 0
    running_loss = 0.0
    pred_scores, true_scores = [], []
    for data in train_dataloader:
        with torch.no_grad():
            if opts.dataset_name[opts.set_id] == 'sysu_sync_diving_3m' or opts.dataset_name[opts.set_id] == 'sysu_sync_diving_10m':
                true_scores.extend(data['true_final_score'].data.numpy())
                true_final_score = data['true_final_score'].cuda()
                true_exec_score = data['true_exec_score'].cuda()
                true_sync_score = data['true_sync_score'].cuda()
                #feat_i3d = data['whole_i3d'].cuda()
                #feat_joints = data['master'].cuda()
            else:
                true_scores.extend(data['score'].data.numpy())
                true_final_score = data['score'].cuda()
                true_exec_score = None
                true_sync_score = None
            feat_i3d = data['whole_i3d'].cuda()
            feat_master = data['master'].cuda()
            feat_slave = data['slave'].cuda()


        pred_exec_score, pred_sync_score, pred_final_score = model_i3d_lstm(feat_i3d, feat_master, feat_slave)
        pred_scores.extend(pred_final_score.data.cpu().numpy())
        loss = criterion(pred_final_score, true_final_score, pred_exec_score, true_exec_score, \
               pred_sync_score, true_sync_score, nn.MSELoss())
        running_loss += loss
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        iter_cnt += 1
    rho, p = stats.spearmanr(pred_scores, true_scores)
    return running_loss/iter_cnt, rho, iter_cnt
        





def test_phase(test_dataloader, criterion):
    with torch.no_grad():
        pred_scores, true_scores = [], []
        loss = 0.0
        iter_cnt = 0
        model_i3d_lstm.eval()
        for data in test_dataloader:
            iter_cnt += 1
            if opts.dataset_name[opts.set_id] == 'sysu_sync_diving_3m' or opts.dataset_name[opts.set_id] == 'sysu_sync_diving_10m':
                true_scores.extend(data['true_final_score'].data.numpy())
                true_final_score = data['true_final_score'].cuda()
                true_exec_score = data['true_exec_score'].cuda()
                true_sync_score = data['true_sync_score'].cuda()
                #feat_i3d = data['whole_i3d'].cuda()
                #feat_joints = data['master'].cuda()
            else:
                true_scores.extend(data['score'].data.numpy())
                true_final_score = data['score'].cuda()
                true_exec_score = None
                true_sync_score = None
            feat_i3d = data['whole_i3d'].cuda()
            feat_master = data['master'].cuda()
            feat_slave = data['slave'].cuda()

            pred_exec_score, pred_sync_score, pred_final_score = model_i3d_lstm(feat_i3d, feat_master, feat_slave)
            pred_scores.extend(pred_final_score.data.cpu().numpy())
            loss += criterion(pred_final_score, true_final_score, pred_exec_score, true_exec_score, \
                    pred_sync_score, true_sync_score, nn.MSELoss())
        rho, p = stats.spearmanr(pred_scores, true_scores)

        return loss/iter_cnt, rho
        #print('pred_final_score: ', pred_scores)
        #print('true_scores: ', true_scores)
        #print('Correlation: ', rho)



def main():
    parameters_2_optimize = (list(model_i3d_lstm.parameters()))

    initial_lr = opts.initial_lr
    weight_decay = opts.weight_decay
    weight_decay = float(sys.argv[2])
    criterion = My_loss(opts._theta, opts.dataset_name[opts.set_id],isPearson=opts.isPearson,isSpearmanr=opts.isSpearmanr)
    test_rho_ls = []
    train_dataset = VideoDataset('train', opts.dataset_name[opts.set_id])
    test_dataset = VideoDataset('test', opts.dataset_name[opts.set_id])
    train_dataloader = DataLoader(train_dataset, batch_size=opts._BATCH_SIZE, shuffle=True, num_workers = 4)
    test_dataloader = DataLoader(test_dataset, batch_size=opts._BATCH_SIZE, shuffle=False, num_workers = 4)

    for epoch in range(opts._EPOCH_NUM):
        optimizer = optim.Adam(parameters_2_optimize, lr=initial_lr, weight_decay=weight_decay)
        train_loss, train_rho, iter_num = train_phase(train_dataloader, optimizer, criterion, epoch)
        if epoch % opts.lr_decay_points[2] == opts.lr_decay_points[0]-1 and opts.lr_loop == True:
            initial_lr *= opts.lr_decay
        elif epoch % opts.lr_decay_points[2] == opts.lr_decay_points[1]-1 and opts.lr_loop == True:
            initial_lr *= opts.lr_decay
        elif epoch % opts.lr_decay_points[2] == opts.lr_decay_points[2]-1 and opts.lr_loop == True:
            initial_lr = opts.initial_lr
        
        if (epoch+1) %2 == 0:
            test_loss, test_rho = test_phase(test_dataloader, criterion)
            print('Dataset: ', opts.dataset_name[opts.set_id])
            print('-----------------------Training------------------------------')
            print('Epoch:', epoch+1, 'Iter: ', iter_num*(epoch+1))
            print('Loss: ',train_loss.data.cpu().numpy())
            print('Correlation: ', train_rho)
            print('')
            print('------------------------Testing------------------------------')
            print('Loss: ',test_loss.data.cpu().numpy())
            print('Correlation: ', test_rho)
            print('')
            print('')
            print('')
        #    if (epoch+1) > opts._EPOCH_NUM - 1000:
        #        test_rho_ls.append(test_rho)
        '''
            writer.add_scalar('train/Loss', train_loss.data.cpu().numpy(), iter_num*(epoch+1))
            writer.add_scalar('train/Corr.', train_rho, iter_num*(epoch+1))
            writer.add_scalar('test/Loss', test_loss.data.cpu().numpy(), iter_num*(epoch+1))
            writer.add_scalar('test/Corr.', test_rho, iter_num*(epoch+1))
        '''
    test_loss, test_rho = test_phase(test_dataloader, criterion)
    print('Dataset: ', opts.dataset_name[opts.set_id])
    print('-----------------------Training------------------------------')
    print('Epoch:', epoch+1, 'Iter: ', iter_num*(epoch+1))
    print('Loss: ',train_loss.data.cpu().numpy())
    print('Correlation: ', train_rho)
    print('')
    print('------------------------Testing------------------------------')
    print('Loss: ',test_loss.data.cpu().numpy())
    print('Correlation: ', test_rho)
    print('')
    print('')
    print('')
    return test_rho



if __name__ == '__main__':
    #writer = SummaryWriter(opts._tb_path+dataset_name[set_id]+'_'+str(randomseed)+'_'+str(datetime.date.today()))
    metric = []
    for _time in range(1):
        #writer = SummaryWriter(opts._tb_path+dataset_name[set_id]+'_'+str(randomseed)+'_'+str(datetime.date.today())+'_'+str(_time))
        model_i3d_lstm = AIM_inserted(opts.batch_normalization)
        model_i3d_lstm = model_i3d_lstm.cuda()
        print('time no.: ', _time)
        metric.append(main())
    if not os.path.isdir(opts._tb_path):
        os.makedirs(opts._tb_path)
    f = open(opts._tb_path+dataset_name[set_id]+'_'+str(randomseed)+'_'+str(datetime.date.today())+'_'+str(sys.argv[4])+'_final.txt','w')
    f.write("dataset: "+dataset_name[set_id]+", weight_decay: "+str(sys.argv[2])+", dropout: "+str(sys.argv[3])+", seed: "+\
        str(randomseed)+', Epoch_num: '+str(opts._EPOCH_NUM)+'lr_decay_points: '+str(opts.lr_decay_points)+'\n final mean-spearmanr: '+str(metric)+\
        ',\n max:'+str(np.max(metric))+',\n mean: '+str(np.mean(metric)))
    f.close()
    print('final mean-spearmanr: %.4f, max: ' % np.mean(metric), np.max(metric))


