import torch
import torch.nn as nn
import opts
import scipy.stats as stats

torch.manual_seed(opts.randomseed); torch.cuda.manual_seed_all(opts.randomseed)

class My_loss(nn.Module):
    def __init__(self, theta, datasetName='SYSU_SyncDiving', isPearson=False, isSpearmanr=False):
        super(My_loss, self).__init__()
        self.theta = theta
        self.datasetName = datasetName
        self.isPearson = isPearson
        self.isSpearmanr = isSpearmanr


    def forward(self, pred_final, true_final, pred_exec=None, true_exec=None, pred_sync=None, true_sync=None, mse=nn.MSELoss()):
        if self.datasetName == 'sysu_sync_diving_3m' or self.datasetName == 'sysu_sync_diving_10m':
            loss = mse(pred_final, true_final) + self.theta * mse(pred_exec, true_exec) \
                    + (1 - self.theta) * mse(pred_sync, true_sync)
        else:
            loss = mse(pred_final, true_final)
            if self.isPearson == True:
                vp = pred_final - torch.mean(pred_final)
                vt = true_final - torch.mean(true_final)
                loss = 1 - torch.sum(vp * vt) / (torch.sqrt(torch.sum(vp ** 2)) * torch.sqrt(torch.sum(vt ** 2)))
            elif self.isSpearmanr == True:
                loss, _s = 100/(1+torch.tensor(stats.spearmanr(pred_final, true_final))) - 50

        return loss
        