import torch
import torch.nn as nn
import opts
import sys

torch.manual_seed(opts.randomseed); torch.cuda.manual_seed_all(opts.randomseed)

class AIM_inserted(nn.Module):
    def __init__(self, batch_normalization = False):
        super(AIM_inserted, self).__init__()
        self.batch_normalization = batch_normalization
        self.decode_left = nn.Linear(opts._FEAT_SIZE, opts._DECODE_SIZE)
        self.att_value = nn.Linear(opts._FEAT_SIZE, opts._DECODE_SIZE)
        self.dropout = nn.Dropout(p=float(sys.argv[3]))
        # self.slave_weight = nn.Parameter(torch.FloatTensor(1), requires_grad=True)
        # self.slave_weight.data.fill_(0.5)
        self.slave_trans = nn.Linear(opts._FRAME_SIZE*2*opts._JOINT_NUM, opts._FRAME_SIZE*2*opts._JOINT_NUM)
        self.rnn = nn.LSTM(opts._FRAME_SIZE*4*opts._JOINT_NUM, opts._DECODE_SIZE, 1, batch_first=True)
        
        self.fc1_left = nn.Sequential(nn.Linear(opts._DECODE_SIZE, opts._FC_SIZE), nn.ReLU(True))
        if batch_normalization == True:
            self.BN1_left = nn.BatchNorm1d(opts._SEG, momentum=opts.bn_momentum)
        self.fc2_left = nn.Linear(opts._FC_SIZE, 1)
        
        self.fc1_mid = nn.Sequential(nn.Linear(opts._DECODE_SIZE, opts._FC_SIZE), nn.ReLU(True))
        if batch_normalization == True:
            self.BN1_mid = nn.BatchNorm1d(opts._FC_SIZE, momentum=opts.bn_momentum)
        self.fc2_mid = nn.Linear(opts._FC_SIZE, 1)
        
        self.fc1_right = nn.Sequential(nn.Linear(opts._DECODE_SIZE, opts._FC_SIZE), nn.ReLU(True))
        if batch_normalization == True:
            self.BN1_right = nn.BatchNorm1d(opts._SEG, momentum=opts.bn_momentum)
        self.fc2_right = nn.Linear(opts._FC_SIZE, 1)


    def forward(self, i3d_feat, master_feat, slave_feat):
        #print(i3d_feat.shape)
        #print(master_feat.shape)
        h_left = self.dropout(i3d_feat)
        #print(i3d_feat.size())
        h_left = self.decode_left(h_left)

        state = None
        #print(master_feat.size())
        #h_right = torch.cat((master_feat - self.slave_weight*slave_feat, master_feat), 2)
        h_right = torch.cat((master_feat - self.slave_trans(slave_feat), master_feat), 2)
        #print(h_right.size())
        h_right = self.dropout(h_right)
        #print(h_right.size())
        h_right, state = self.rnn(h_right)

        h_att = self.dropout(i3d_feat)
        h_att = self.att_value(h_att)
        #print(h_left.shape)
        #print(h_att.shape)
        #print((h_left * h_att).shape)
        #print((h_left * h_att).sum(2).shape)

        w_att = nn.functional.softmax(torch.cat(((h_left * h_att).sum(2).view([-1, opts._SEG, 1]), 
        	                                     (h_right * h_att).sum(2).view([-1, opts._SEG, 1])), 2), 2)
        h_att = w_att[:,:,0:1] * h_left + w_att[:,:,1:2] * h_right

        h_left = self.dropout(h_left)
        h_left = self.fc1_left(h_left)
        if self.batch_normalization == True:
            h_left = self.BN1_left(h_left)
        h_left = self.dropout(h_left)
        h_left = self.fc2_left(h_left)

        h_right = self.dropout(h_right)
        h_right = self.fc1_right(h_right)
        if self.batch_normalization == True:
            h_right = self.BN1_left(h_right)
        h_right = self.dropout(h_right)
        h_right = self.fc2_right(h_right)

        h_att = self.dropout(h_att)
        h_att = self.fc1_mid(h_att)
        if self.batch_normalization == True:
            h_att = self.BN1_left(h_att)
        h_att = self.dropout(h_att)
        h_att = self.fc2_mid(h_att)

        #print(h_left.size(), h_right.size(), h_att.size())
        pred_exec_score = h_left.squeeze().sum(1)
        pred_sync_score = h_right.squeeze().sum(1)
        pred_final_score = h_att.squeeze().sum(1)
        #print(pred_final_score.size())
        return pred_exec_score, pred_sync_score, pred_final_score